/***************************************************************************
 *                          Zanoza Modeler v3.x                            *
 *      This is unpublished propietary source code of Zanoza Software.     *
 *      The copyright notice does not evidence any actual or intended      *
 *                    publication of such source code.                     *
 *                                                                         *
 *                 Copyright (c) 2002-2012 Zanoza Software.                *
 *                                                                         *
 *                        All rights reserved.                             *
 ***************************************************************************
 *--------------------------------------------------------------------------
 * @author:   Oleg M.
 *--------------------------------------------------------------------------
 * @purpose:  ZModeler 3.x Color + Bump + Diffuse + Spec + EnvMap shader
 *--------------------------------------------------------------------------
 * @histroy: 29.08.2011. Created
 * @histroy: 31.08.2012. Damage support added
<[[
UserName = "Bump+Diff+Spec+Env"
Description = "Color, Diffise, Bump, Specular and Env map"
Version  = 1.7
Options  = SOLID BUMP SPECULAR ENVIRONMENT FRESNEL TRANSITION DUSTCOLOR DIRTCOLOR SCRATCHCOLOR BURNCOLOR DIFFUSECOLOR2 DIFFUSECOLOR3 DIFFUSECOLOR4
Samplers = BUMPMAP DETAIL SPECMAP ENVMAP
ManualUV1 = v.vUV01.xy
ManualUV2 = v.vUV01.zw
ManualUV3 = v.vUV23.xy
ManualUV4 = v.vUV23.zw
EnvUV = vReflect.xy
WorldUV = v.vUVRefl.zw
AlphaRegister = 29

[BUMPMAP]
Variable = g_BumpMap
Register = 0
RegisterTrans = 1
Mask = RGB AGB GRB
Scale = 33.R
Level = 30.R
AngularLevel = 30.G

[DETAIL]
Variable = g_Diffuse
Register = 2
RegisterTrans = 3
Mask = R G B A RGB
Scale = 33.G
Level = 30.B
AngularLevel = 30.A

[SPECMAP]
Variable = g_SpecMap
Register = 4
RegisterTrans = 5
Mask = R G B A RGB
Scale = 33.B
Level = 31.R
AngularLevel = 31.G

[ENVMAP]
Variable = g_Refl
Register = 6
Mask = RGB
Scale = 33.A
Level = 32.R
AngularLevel = 32.B

[DUSTCOLOR]
Register    = 34

[DIRTCOLOR]
Register    = 35

[SCRATCHCOLOR]
Register    = 36

[BURNCOLOR]
Register    = 37

[DIFFUSECOLOR2]
Register    = 38

[DIFFUSECOLOR3]
Register    = 39

[DIFFUSECOLOR4]
Register    = 40

[CUSTOM.1]
Name = Opacity Amplifier
Slot = 32.G
Type = float
Min = 0.0
Max = 1.0
Default = 0.0

[CUSTOM.2]
Name = Fresnel
Slot = 32.A
Type = float
Min = 0.001
Max = 0.3
Default = 0.2

[ALPHABLEND]

[ALPHATEST]


]]>
*/

#pragma pack_matrix( row_major ) 

struct VS_INPUT
{
    float4 vPosition        : POSITION;
    float4 vNormal          : NORMAL;
    float4 vTangent         : TANGENT;
    float4 vColorDif        : COLOR0;
    float4 vUV01            : TEXCOORD0;
    float4 vUV23            : TEXCOORD1;
    float4 vDeformPosition  : TEXCOORD5;
    float4 vDeformNormal    : COLOR2;
    uint4  vBones           : BLENDINDICES;
    float4 vWeights         : BLENDWEIGHT;
};

struct VS_OUTPUT
{
    float4 vPosition  : POSITION;   //transformed position
    float4 vColor     : COLOR0;     //vertex color composition {ambient, diffuse, specular, 1}
    float4 vUV01      : TEXCOORD0;
    float4 vUV23      : TEXCOORD1;
    float4 vUVRefl    : TEXCOORD2;  //environment reflection;
    //precomputed stuff:
    float4 vNormal    : TEXCOORD4;  //world-space normal;
    float4 vViewer    : TEXCOORD5;  //camera-space position;
    float4 vBumpLit   : TEXCOORD6;  //tangent-space light;
    float4 vBumpViewer: TEXCOORD7;  //tangent-space position;
};

//----------------------------------------------------------------------
// globals:
//----------------------------------------------------------------------

#include "zmCommon.inl"

float4      g_alpha          : register(c29);  //[minlevel:testvalue, maxlevel, mapaffect, angular level]
float4      g_textureAmp     : register(c30);  //[bump level, bump angular level, diff level, diff angular level]
float4      g_specAmp        : register(c31);  //[spec level, spec angular level, 0, 0]
float4      g_environment    : register(c32);  //[level, opacity strengthening, angular level, fresnel value]
float4      g_textureScale   : register(c33);  //[bump uv scale, diff uv scale, spec uv scale, refl uv scale]
float4      g_dustColor      : register(c34);  //dust color r,g,b and level in .a
float4      g_dirtColor      : register(c35);  //dirt color r,g,b and level in .a
float4      g_scratchColor   : register(c36);  //scratch color r,g,b and level in .a
float4      g_burnColor      : register(c37);  //burn color r,g,b and level in .a
                                               //level of main diffuse is in g_colors[1].a;
float4      g_diffuse2Color  : register(c38);  //diffuse2 color r,g,b and level in .a
float4      g_diffuse3Color  : register(c39);  //diffuse3 color r,g,b and level in .a
float4      g_diffuse4Color  : register(c40);  //diffuse4 color r,g,b and level in .a

// textures:
sampler     g_BumpMap        : register(s[0]);
sampler     g_BumpMapTrans   : register(s[1]);
sampler     g_Diffuse        : register(s[2]);
sampler     g_DiffuseTrans   : register(s[3]);
sampler     g_SpecMap        : register(s[4]);
sampler     g_SpecMapTrans   : register(s[5]);
sampler     g_Refl           : register(s[6]);

#ifdef VERTEX_SHADER
VS_OUTPUT mainVS(in VS_INPUT v)
{
    VS_OUTPUT output = (VS_OUTPUT)0;
    output.vNormal = v.vNormal;
    float4 vTangent = v.vTangent;
    output.vPosition = transformInput(v, output.vNormal, vTangent, output.vViewer);
    output.vColor = any(v.vColorDif.rgb) ? v.vColorDif : 1;

    output.vUV01 = v.vUV01;
    output.vUV23 = v.vUV23;
    //bump-map reflection can mix this spehere-env uv mapping with bump-reflected env;
    output.vUVRefl.xy = (mul(output.vNormal.xyz, (float3x3)g_mView).xy+1)/2;
    output.vUVRefl.zw = v.vPosition.xz;

    // bump-transformed light:
    //-----
    float3x3 worldToTangentSpace;
    worldToTangentSpace[0] = vTangent.xyz;
    worldToTangentSpace[1] = normalize(cross(vTangent.xyz, output.vNormal.xyz)*(-v.vTangent.w));
    worldToTangentSpace[2] = output.vNormal.xyz;
    output.vBumpLit = normalize(float4(mul(worldToTangentSpace, -g_lightDir.xyz), 1));
    output.vBumpViewer = float4(mul(worldToTangentSpace, output.vViewer.xyz), 1);

    return output;
}
#endif //VERTEX_SHADER

#ifdef PIXEL_SHADER
//------------------------------------------------------------
//  PixelShader: generic Dif+Bump+SpecMap+Refl;
//------------------------------------------------------------
float4 mainPS(in VS_OUTPUT v) : COLOR0
{
    float4 oColor;
    float4 CONST_TEXEL   = float4(1,1,1,1);
    float4 VCOLOR_TEXEL  = v.vColor;
    float2 vReflect      = v.vUVRefl.xy;
    float4 BUMPMAP_TEXEL = tex2D(BUMPMAP_SOURCE, BUMPMAP_UV_SOURCE/g_textureScale.x);
    float4 DETAIL_TEXEL  = tex2D(DETAIL_SOURCE, DETAIL_UV_SOURCE/g_textureScale.y);
    float4 SPECMAP_TEXEL = tex2D(SPECMAP_SOURCE, SPECMAP_UV_SOURCE/g_textureScale.z);
    float4 ENVMAP_TEXEL  = CONST_TEXEL;
    DETAIL_TEXEL  = lerp(DETAIL_TEXEL,  tex2D(g_DiffuseTrans, DETAIL_UV_SOURCE/g_textureScale.y), DETAIL_TRANSITION);
    SPECMAP_TEXEL = lerp(SPECMAP_TEXEL, tex2D(g_SpecMapTrans, SPECMAP_UV_SOURCE/g_textureScale.z), SPECMAP_TRANSITION);
    BUMPMAP_TEXEL = lerp(BUMPMAP_TEXEL, tex2D(g_BumpMapTrans, BUMPMAP_UV_SOURCE/g_textureScale.x), BUMPMAP_TRANSITION);

    float3 normal = normalize(v.vNormal.xyz);
    float3 bumpNormal = normalize((2*(BUMPMAP_INPUT*BUMPMAP_INPUT_MASK)BUMPMAP_MASK)-1);
    float3 bumpLight = normalize(v.vBumpLit.xyz);
    
    //
    // regular vs bump-mapped normals and light:
    // float4 LightTemp is [non-bump diff, non-bump spec, bump-diff, bump spec]
    float4 specData = 0;
#ifdef SPECULAR_ENABLED
    specData.xy = float2(dot(-g_lightDir.xyz, normal), dot(normal, normalize(v.vViewer.xyz-g_lightDir.xyz)));
    specData.zw = float2(dot(bumpLight, bumpNormal), dot(bumpNormal, normalize(bumpLight+v.vBumpViewer.xyz)));
    float4 LightTemp = float4(
                    saturate(specData.x), 
                    lit(specData.x, specData.y, g_colors[2].a).z,
                    saturate(specData.z),
                    lit(specData.z, specData.w, g_colors[2].a).z); 
    
#else
    float4 LightTemp = float4(
                    saturate(dot(-g_lightDir.xyz, normal)), 
                    0,
                    saturate(dot(bumpLight, bumpNormal)),
                    0); 
#endif //SPECULAR_ENABLED
    // alter diffuse light factor by ambient illumitaion addon:
    LightTemp.x = saturate(LightTemp.x + (g_ambientLight.r + g_ambientLight.g + g_ambientLight.b)/3);
    LightTemp.z = saturate(LightTemp.z + (g_ambientLight.r + g_ambientLight.g + g_ambientLight.b)/3);

    //
    // angular level correction of bumpmap
    oColor.a = lerp(saturate(dot(v.vViewer.xyz, normal)), saturate(dot(v.vBumpViewer.xyz, bumpNormal)), g_textureAmp.x);
    
   
    //
    // angular level on bump:
    if (g_textureAmp.y >= 0)
      specData.x = lerp(1, oColor.a, g_textureAmp.y)*g_textureAmp.x;
    else
      specData.x = lerp(1, 1-oColor.a, abs(g_textureAmp.y))*g_textureAmp.x;

    //
    // environment UV computation are wrong in tangent space (vBumpReflect); 
    // however it's suitable at the moment, since I reduce affection (specData.x)
    // when bump-map normal pointing around (0,0,1)
    //
    float2 vBumpReflect   = (reflect(bumpNormal, v.vBumpViewer.xyz).xy + 1)/2;
    vReflect              = lerp(vReflect, vBumpReflect, specData.x*(1-saturate(bumpNormal.z)));
            
    ENVMAP_TEXEL          = tex2D(ENVMAP_SOURCE, ENVMAP_UV_SOURCE*g_textureScale.w);
    
    float3 fEnvironColor = (ENVMAP_INPUT*(ENVMAP_INPUT_MASK))ENVMAP_MASK;
    float3 fDetailColor    = (DETAIL_INPUT)DETAIL_MASK;
    float3 fSpecColor    = (SPECMAP_INPUT)SPECMAP_MASK;
     
    //
    // angular level on diffuse:
    if (g_textureAmp.w >= 0)
      specData.y = lerp(1, oColor.a, g_textureAmp.w)*g_textureAmp.z*(DETAIL_INPUT_MASK);
    else
      specData.y = lerp(1, 1-oColor.a, abs(g_textureAmp.w))*g_textureAmp.z*(DETAIL_INPUT_MASK);

    fDetailColor = lerp(1, fDetailColor, specData.y);
    //
    // angular level on spec:
    if (g_specAmp.y >= 0)
      specData.z = lerp(1, oColor.a, g_specAmp.y)*g_specAmp.x;
    else
      specData.z = lerp(1, 1-oColor.a, abs(g_specAmp.y))*g_specAmp.x;
     
    //
    // angular level on env:
    if (g_environment.z >= 0)
      specData.w = lerp(1, oColor.a, g_environment.z)*g_environment.x;
    else
      specData.w = lerp(1, 1-oColor.a, abs(g_environment.z))*g_environment.x;

    float4 lightInfo;
    lightInfo.g = (1-oColor.a);
    lightInfo.r = lightInfo.g*lightInfo.g;
    lightInfo.r = lightInfo.r*lightInfo.r;
    lightInfo.r = saturate(mad(lightInfo.g*lightInfo.r, 1-saturate(g_environment.w), g_environment.w));
    
    fEnvironColor = fEnvironColor*saturate(specData.w+(0.5 + g_environment.x)*lightInfo.r);

    // diffuse corrected on bump:
    lightInfo.b = lerp(LightTemp.x, LightTemp.z, specData.x);
    // specular corrected on bump:
    lightInfo.a = lerp(LightTemp.y, LightTemp.w, specData.x);

    //
    // advanced diffuse color, and shine/opacity overload controllers:
    ADVANCED_COLOR_DIFFUSELEVEL(fAdvColorLevels)
    ADVANCED_COLOR_DIFFUSEVAR(fAdvancedDiffuseColor, fAdvColorLevels)
    ADVANCED_COLOR_EFFECTLEVELS(fEffectLevels)
    ADVANCED_COLOR_EFFECTVAR(fEffectColor, fEffectLevels)
    ADVANCED_COLOR_OPTIONS(advShineOpac, fEffectLevels, fAdvColorLevels)
    
    //regular:
    oColor.rgb =
                // ambient: 
                (g_colors[0].rgb + 0.2*v.vColor.rgb + 0.8*fDetailColor*g_ambientLight.rgb)*g_ambientLight.rgb +
                // diffuse component:
                (fAdvancedDiffuseColor*v.vColor.rgb*fDetailColor*(1-fEffectColor.a) + fEffectColor.rgb*fEffectColor.a)*max(g_colors[3].a, lightInfo.b) +
                // matted:
                advShineOpac.r*(
                  // specular component:
                  g_colors[2].rgb*(fSpecColor*(SPECMAP_INPUT_MASK)*specData.z)*(lightInfo.a + lightInfo.a*(1+lightInfo.b)*saturate(advShineOpac.g)) +
                  // reflection component:
                  fEnvironColor + 
                  // emissive component:
                  g_colors[3].rgb*g_colors[3].a
                );
                

#ifdef ALPHABLEND_ENABLED
    if (g_alpha.w >= 0)
      oColor.a = lerp(1, oColor.a, g_alpha.w);
    else
      oColor.a = lerp(1, 1-oColor.a, abs(g_alpha.w));
    oColor.a = saturate(lerp(1, ALPHABLEND_INPUT, g_alpha.z)*ALPHABLEND_VERTEXCOLOR*oColor.a*g_alpha.y + 
                        length(fEnvironColor)*g_environment.y + advShineOpac.b);
    
#else
    oColor.a = g_alpha.y;
#endif//ALPHABLEND_ENABLED
#ifdef ALPHATEST_ENABLED
    clip(oColor.a - g_alpha.x);
#endif//ALPHATEST_ENABLED

    return oColor;
}
#endif //PIXEL_SHADER

